from typing import Hashable

from ortools.sat.python.cp_model import IntVar

from .cp_model_with_fixed_interval import CpModelWithFixedInterval


class CpModelWithOptionalFixedInterval(CpModelWithFixedInterval):
    # Variables

    var_op_is_present: dict[Hashable, IntVar]
    """
    Dictionary to store presence indicator variables for each operation in a job.
    """

    var_op_intvl_opt_fixed_name_set: set[str]
    """
    Set to store names of optional fixed interval variables.
    """

    def __init__(self, horizon: int):
        """Initialize the CpModelWithOptionalFixedInterval class.

        Args:
            horizon (int): The horizon for the scheduling problem,
                           which is the maximum time that any operation can end.
        """
        super().__init__(horizon)

        # Parameter
        self.horizon = horizon

        # Initialize dictionaries to store variables
        self.var_op_is_present = {}

        # Initialize the set to store interval variable names
        self.var_op_intvl_opt_fixed_name_set = set()

    def define_optional_fixed_interval_var(
        self, job_idx: str, stage_idx: str, mc_idx: str, processing_time: int
    ):
        """Define an optional interval variable with fixed processing time
        for an operation.

        Args:
            job_idx (str): The index of the job.
            stage_idx (str): The index of the stage.
            mc_idx (str): The index of the machine.
            processing_time (int): The processing time of the operation.
        """
        # method var_optional_casts on line 184 in constraint_program_model.py
        suffix = f"{job_idx}_{stage_idx}_{mc_idx}"
        var_intvl_name = f"intvl_opt_fixed_{suffix}"
        # Check if the interval variable name already exists
        if var_intvl_name in self.var_op_intvl_opt_fixed_name_set:
            raise ValueError(
                f"Interval variable name '{var_intvl_name}' already exists."
            )

        # Create an optional interval variable with the specified processing time
        start_var = self.new_int_var(0, self.horizon, f"start_{suffix}")
        end_var = self.new_int_var(0, self.horizon, f"end_{suffix}")
        is_present_var = self.new_bool_var(f"is_present_{suffix}")
        intvl_var = self.new_optional_interval_var(
            start_var, processing_time, end_var, is_present_var, var_intvl_name
        )

        # Add each variable to the corresponding dictionaries
        self.var_op_start[job_idx, stage_idx, mc_idx] = start_var
        self.var_op_end[job_idx, stage_idx, mc_idx] = end_var
        self.var_op_is_present[job_idx, stage_idx, mc_idx] = is_present_var
        self.var_op_intvl[job_idx, stage_idx, mc_idx] = intvl_var

        # Add name of the new variable to the set
        self.var_op_intvl_opt_fixed_name_set.add(var_intvl_name)



# tests/fake_cp_model.py
from typing import Optional

from routix import ElapsedTimer

from src.mbls.cpsat.custom_cp_model import CustomCpModel


class FakeCpModel(CustomCpModel):
    def __init__(self, maximize=False):
        super().__init__()
        self._maximize = maximize
        self._progress_log = [
            (0.0, 100.0, 60.0),
            (1.0, 90.0, 70.0),
            (2.0, 85.0, 50.0),
        ]

    def solve_with_prog_logger(
        self,
        computational_time: float,
        num_workers: int,
        random_seed: Optional[int] = None,
        timer: Optional[ElapsedTimer] = None,
    ):
        return "OPTIMAL", 3.0, self._progress_log[-1][1], self._progress_log[-1][2]

    def get_progress_log(self):
        return self._progress_log

    def is_maximize(self) -> bool:
        return self._maximize

    def set_num_base_constraints(self):
        pass



import pytest

from mbls.cpsat.cp_model_with_optional_fixed_interval import (
    CpModelWithOptionalFixedInterval,
)


@pytest.fixture
def model():
    """Fixture to create a CpModelWithOptionalInterval instance."""
    return CpModelWithOptionalFixedInterval(horizon=100)


def test_initialization(model: CpModelWithOptionalFixedInterval):
    """Test that the model initializes correctly."""
    assert model.horizon == 100
    assert isinstance(model.var_op_start, dict)
    assert isinstance(model.var_op_end, dict)
    assert isinstance(model.var_op_is_present, dict)
    assert isinstance(model.var_op_intvl, dict)


def test_define_optional_fixed_interval_var(model: CpModelWithOptionalFixedInterval):
    """Test the define_optional_interval_var method."""
    job_idx = "job1"
    stage_idx = "stage1"
    mc_idx = "mc1"
    processing_time = 10

    model.define_optional_fixed_interval_var(
        job_idx, stage_idx, mc_idx, processing_time
    )

    # Check that variables are created and stored correctly
    assert (job_idx, stage_idx, mc_idx) in model.var_op_start

    start_var = model.var_op_start[job_idx, stage_idx, mc_idx]
    end_var = model.var_op_end[job_idx, stage_idx, mc_idx]
    is_present_var = model.var_op_is_present[job_idx, stage_idx, mc_idx]
    interval_var = model.var_op_intvl[job_idx, stage_idx, mc_idx]

    assert start_var is not None
    assert end_var is not None
    assert is_present_var is not None
    assert interval_var is not None

    # Check variable properties
    assert start_var.Name() == "start_job1_stage1_mc1"
    assert end_var.Name() == "end_job1_stage1_mc1"
    assert is_present_var.Name() == "is_present_job1_stage1_mc1"
    assert interval_var.Name() == "intvl_opt_fixed_job1_stage1_mc1"


def test_multiple_operations(model: CpModelWithOptionalFixedInterval):
    """Test adding multiple operations to the model."""
    operations = [
        ("job1", "stage1", "mc1", 10),
        ("job1", "stage2", "mc2", 15),
        ("job2", "stage1", "mc1", 20),
    ]

    for job_idx, stage_idx, mc_idx, processing_time in operations:
        model.define_optional_fixed_interval_var(
            job_idx, stage_idx, mc_idx, processing_time
        )

    # Check that all operations are added correctly
    for job_idx, stage_idx, mc_idx, _ in operations:
        assert (job_idx, stage_idx, mc_idx) in model.var_op_start
        assert (job_idx, stage_idx, mc_idx) in model.var_op_end
        assert (job_idx, stage_idx, mc_idx) in model.var_op_is_present
        assert (job_idx, stage_idx, mc_idx) in model.var_op_intvl



from abc import abstractmethod
from datetime import datetime
from typing import Any, Generic, Optional, TypeVar

from routix import (
    DynamicDataObject,
    ElapsedTimer,
    SolverOutputSummary,
    SubroutineController,
)

from .custom_cp_model import CustomCpModel
from .obj_value_bound_store import ObjValueBoundStore

ProblemT = TypeVar("ProblemT")  # Type for the problem instance
CpModelT = TypeVar("CpModelT", bound=CustomCpModel)
StoppingCriteriaT = TypeVar("StoppingCriteriaT", bound=DynamicDataObject)


class CpSubroutineController(
    SubroutineController, Generic[ProblemT, CpModelT, StoppingCriteriaT]
):
    """
    Subroutine controller utilizing OR-Tools CP model.
    This controller manages the execution of a CP model subroutine,
    including the creation of the model, solving it, and logging results.
    """

    def __init__(
        self,
        instance: ProblemT,
        shared_param_dict: dict,
        cp_model_class: type[CpModelT],
        subroutine_flow: DynamicDataObject,
        stopping_criteria: StoppingCriteriaT,
        start_dt: Optional[datetime] = None,
        expr_name: Optional[str] = None,
    ):
        """
        Initialize the CpSubroutineController.
        - Set up the subroutine controller with algorithm data.
        - Create an instance of base CP model using the provided instance and shared parameters.
        - Set the number of base constraints of the CP model.

        Args:
            instance (ProblemT): Instance-specific parameters for the CP model.
            shared_param_dict (dict): Shared parameters for the CP model.
            cp_model_class (type[CpModelT]): The class of the CP model to be used.
            subroutine_flow (DynamicDataObject): The flow of the subroutine to be executed.
            stopping_criteria (StoppingCriteriaT): Stopping criteria for the controller.
            start_dt (Optional[datetime], optional): Start date and time for the controller. Defaults to None.
            expr_name (Optional[str], optional): Name of the experiment. Defaults to None.
        """
        _expr_name = expr_name or str(
            getattr(instance, "name", "CP Subroutine Controller")
        )
        super().__init__(
            name=_expr_name,
            subroutine_flow=subroutine_flow,
            stopping_criteria=stopping_criteria,
            start_dt=start_dt,
        )

        self.instance = instance
        """Instance-specific parameters for the CP model."""
        self.shared_param_dict = shared_param_dict
        """Shared parameters for the CP model."""
        self.cp_model_class = cp_model_class
        """The class of the CP model to be used."""

        self.obj_store = ObjValueBoundStore[float]()
        """Store for objective value and bound time series."""

        self.cp_model = self.create_base_cp_model()
        self.cp_model.set_num_base_constraints()

    @property
    def obj_value_log(self) -> list[tuple[float, float]]:
        """Get the objective value log.

        Returns:
            list[tuple[float, float]]: List of tuples containing (elapsed time, objective value).
        """
        return self.obj_store.obj_value_series.items()

    @property
    def obj_bound_log(self) -> list[tuple[float, float]]:
        """Get the objective bound log.

        Returns:
            list[tuple[float, float]]: List of tuples containing (elapsed time, objective bound).
        """
        return self.obj_store.obj_bound_series.items()

    @abstractmethod
    def create_base_cp_model(self) -> CpModelT:
        """Create and return a CP model instance."""
        pass

    def add_obj_log(
        self,
        log_entry: tuple[float, float, float],
        is_maximize: bool = False,
        obj_value_is_valid: bool = False,
        obj_bound_is_valid: bool = False,
    ) -> None:
        """Update the objective value and bound logs with a single tuple.

        Args:
            log_entry (tuple[float, float, float]): (elapsed time, objective value, objective bound)
            is_maximize (bool, optional): If True, indicates maximization problem.
                Defaults to False.
            obj_value_is_valid (bool, optional): If True, adds the objective value log.
                Defaults to False.
            obj_bound_is_valid (bool, optional): If True, adds the objective bound log.
                Defaults to False.
        """
        if obj_value_is_valid:
            self.obj_store.add_obj_value(
                log_entry[0], log_entry[1], is_maximize=is_maximize
            )
        if obj_bound_is_valid:
            self.obj_store.add_obj_bound(
                log_entry[0], log_entry[2], is_maximize=is_maximize
            )

    def append_obj_log(
        self,
        log_list: list[tuple[float, float, float]],
        is_maximize: bool = False,
        obj_value_is_valid: bool = False,
        obj_bound_is_valid: bool = False,
        last_timestamp_note: Optional[Any] = None,
    ) -> None:
        """Update the objective value and bound logs with a list of tuples.

        Args:
            log_list (list[tuple[float, float, float]]): List of tuples containing
                (elapsed time, objective value, objective bound).
            is_maximize (bool, optional): If True, indicates maximization problem.
                Defaults to False.
            obj_value_is_valid (bool, optional): If True, adds the objective value log.
                Defaults to False.
            obj_bound_is_valid (bool, optional): If True, adds the objective bound log.
                Defaults to False.
            last_timestamp_note (Optional[Any], optional): Optional note for the last timestamp.
                If not provided, uses the call context of the current method.
        """
        for log_entry in log_list:
            self.add_obj_log(
                log_entry,
                is_maximize=is_maximize,
                obj_value_is_valid=obj_value_is_valid,
                obj_bound_is_valid=obj_bound_is_valid,
            )
        _last_timestamp_note = (
            last_timestamp_note or self._get_call_context_of_current_method()
        )
        self.obj_store.add_last_timestamp_note(
            _last_timestamp_note,
            obj_value_is_valid=obj_value_is_valid,
            obj_bound_is_valid=obj_bound_is_valid,
        )

    def solve_cp_model(
        self,
        cp_model: CpModelT,
        computational_time: float,
        num_workers: int,
        random_seed: Optional[int] = None,
        timer: Optional[ElapsedTimer] = None,
        obj_value_is_valid: bool = False,
        obj_bound_is_valid: bool = False,
        last_timestamp_note: Optional[Any] = None,
    ) -> SolverOutputSummary:
        """Solve the given CP model.

        Args:
            cp_model (CpModelT): The CP model to be solved.
            computational_time (float): The maximum computational time in seconds.
            num_workers (int): The number of parallel workers (i.e. threads) to use during search.
            random_seed (Optional[int], optional): Random seed for reproducibility. Defaults to None.
            timer (Optional[ElapsedTimer], optional): Timer to be passed to solver callback. Defaults to None.
            obj_value_is_valid (bool, optional): If True, adds the objective value log.
                Defaults to False.
            obj_bound_is_valid (bool, optional): If True, adds the objective bound log.
                Defaults to False.
            last_timestamp_note (Optional[Any], optional): Optional note for the last timestamp.
                Defaults to None.

        Raises:
            AttributeError: If self.cp_model is not initialized.

        Returns:
            SolverOutputSummary: A summary of the solver output, including status,
                elapsed time, objective value, best objective bound, and progress log.
        """
        (solver_status, elapsed_time, obj_value, obj_bound) = (
            cp_model.solve_with_prog_logger(
                computational_time=computational_time,
                num_workers=num_workers,
                random_seed=random_seed,
                timer=timer,
            )
        )
        progress_log = cp_model.get_progress_log()
        self.append_obj_log(
            log_list=progress_log,
            is_maximize=cp_model.is_maximize(),
            obj_value_is_valid=obj_value_is_valid,
            obj_bound_is_valid=obj_bound_is_valid,
            last_timestamp_note=last_timestamp_note,
        )

        return SolverOutputSummary(
            status=solver_status,
            elapsed_time=elapsed_time,
            objective_value=obj_value,
            best_objective_bound=obj_bound,
            progress_log=progress_log,
        )

    def solve_current_cp_model(
        self,
        computational_time: float,
        num_workers: int,
        random_seed: Optional[int] = None,
        timer: Optional[ElapsedTimer] = None,
        obj_value_is_valid: bool = False,
        obj_bound_is_valid: bool = False,
        last_timestamp_note: Optional[Any] = None,
    ) -> SolverOutputSummary:
        """Solve the current CP model.

        Args:
            computational_time (float): The maximum computational time in seconds.
            num_workers (int): The number of parallel workers (i.e. threads) to use during search.
            random_seed (Optional[int], optional): Random seed for reproducibility. Defaults to None.
            timer (Optional[ElapsedTimer], optional): Timer to be passed to solver callback. Defaults to None.
            obj_value_is_valid (bool, optional): If True, adds the objective value log.
                Defaults to False.
            obj_bound_is_valid (bool, optional): If True, adds the objective bound log.
                Defaults to False.
            last_timestamp_note (Optional[Any], optional): Optional note for the last timestamp.
                Defaults to None.

        Raises:
            AttributeError: If self.cp_model is not initialized.

        Returns:
            SolverOutputSummary: A summary of the solver output, including status,
                elapsed time, objective value, best objective bound, and progress log.
        """
        if not hasattr(self, "cp_model"):
            raise AttributeError("CP model is not initialized.")
        return self.solve_cp_model(
            cp_model=self.cp_model,
            computational_time=computational_time,
            num_workers=num_workers,
            random_seed=random_seed,
            timer=timer,
            obj_value_is_valid=obj_value_is_valid,
            obj_bound_is_valid=obj_bound_is_valid,
            last_timestamp_note=last_timestamp_note,
        )



import pytest

from src.mbls import ElapsedTimer, SolverStatus
from src.mbls.cpsat.custom_cp_model import CustomCpModel


@pytest.fixture
def model():
    """Fixture to create a CustomCpModel instance."""
    return CustomCpModel()


def test_init_solver(model: CustomCpModel):
    model.init_solver(computational_time=10.0, num_workers=4)
    assert model.solver.parameters.max_time_in_seconds == 10.0
    assert model.solver.parameters.num_workers == 4


def test_solve_and_get_status(model: CustomCpModel):
    model.init_solver(computational_time=10.0, num_workers=4)
    status, elapsed_time, ub, lb = model.solve_and_get_status(
        computational_time=10.0, num_workers=4
    )
    assert isinstance(status, int)
    assert isinstance(elapsed_time, float)
    assert isinstance(ub, float)
    assert isinstance(lb, float)


def test_change_domain(model: CustomCpModel):
    var = model.new_int_var(0, 10, "test_var")
    model.change_domain(var, [5, 15])
    assert var.Proto().domain == [5, 15]


def test_is_maximize(model: CustomCpModel):
    assert model.is_maximize() is False


def test_set_num_base_constraints(model: CustomCpModel):
    # Create integer variables
    x = model.new_int_var(0, 10, "x")
    y = model.new_int_var(0, 10, "y")

    # Add a constraint
    model.add(x + y <= 15)

    # Set the number of base constraints
    model.set_num_base_constraints()

    # Assert that the number of base constraints is updated correctly
    assert model.num_base_constraints == 1


def test_delete_added_constraints(model: CustomCpModel):
    # Create integer variables
    x = model.new_int_var(0, 10, "x")
    y = model.new_int_var(0, 10, "y")

    # Add base constraints
    model.add(x >= 0)
    model.add(x + y <= 15)
    model.set_num_base_constraints()

    # Add additional constraints
    model.add(x - y >= 5)  # Added constraint 1
    model.add(x + 2 * y == 20)  # Added constraint 2

    # Delete added constraints
    model.delete_added_constraints()

    # Assert that only the base constraints remain
    assert model.get_next_constr_idx() == model.num_base_constraints


def test_solve_with_prog_logger(model: CustomCpModel):
    timer = ElapsedTimer()
    model.init_solver(computational_time=10.0, num_workers=4)

    # Example CP model by Google
    num_vals = 3
    x = model.new_int_var(0, num_vals - 1, "x")
    y = model.new_int_var(0, num_vals - 1, "y")
    z = model.new_int_var(0, num_vals - 1, "z")
    model.add(x != y)

    status, elapsed_time, ub, lb = model.solve_with_prog_logger(
        computational_time=10.0, num_workers=4, timer=timer
    )
    assert isinstance(status, str)
    assert isinstance(elapsed_time, float)
    assert isinstance(ub, float)
    assert isinstance(lb, float)
    if SolverStatus.found_feasible_solution(status):
        print(f"x = {model.solver.value(x)}")
        print(f"y = {model.solver.value(y)}")
        print(f"z = {model.solver.value(z)}")
    else:
        print("No solution found.")



from routix import DynamicDataObject

from src.mbls.cpsat import CpSubroutineController

from .fake_cp_model import FakeCpModel


class DummyProblem:
    name = "test-problem"


class DummyStoppingCriteria(DynamicDataObject):
    pass


class DummyCpController(
    CpSubroutineController[DummyProblem, FakeCpModel, DummyStoppingCriteria]
):
    def is_stopping_condition(self) -> bool:
        return False

    def post_run_process(self):
        pass

    def create_base_cp_model(self) -> FakeCpModel:
        return FakeCpModel(maximize=False)


def test_add_obj_log_minimize():
    controller = DummyCpController(
        instance=DummyProblem(),
        shared_param_dict={},
        cp_model_class=FakeCpModel,
        subroutine_flow=DynamicDataObject({}),
        stopping_criteria=DummyStoppingCriteria({}),
    )

    controller.add_obj_log(
        (0.0, 100.0, 150.0),
        is_maximize=False,
        obj_value_is_valid=True,
        obj_bound_is_valid=True,
    )
    controller.add_obj_log(
        (1.0, 95.0, 140.0),
        is_maximize=False,
        obj_value_is_valid=True,
        obj_bound_is_valid=True,
    )
    controller.add_obj_log(
        (2.0, 105.0, 160.0),
        is_maximize=False,
        obj_value_is_valid=True,
        obj_bound_is_valid=True,
    )

    val_log = controller.obj_value_log
    bound_log = controller.obj_bound_log

    assert val_log == [(0.0, 100.0), (1.0, 95.0)]
    assert bound_log == [(0.0, 150.0), (2.0, 160.0)]


def test_append_obj_log_maximize():
    class MaxController(DummyCpController):
        def create_base_cp_model(self) -> FakeCpModel:
            return FakeCpModel(maximize=True)

    controller = MaxController(
        instance=DummyProblem(),
        shared_param_dict={},
        cp_model_class=FakeCpModel,
        subroutine_flow=DynamicDataObject({}),
        stopping_criteria=DummyStoppingCriteria({}),
    )

    logs = [
        (0.0, 10.0, 20.0),
        (1.0, 15.0, 18.0),
        (2.0, 13.0, 22.0),
    ]
    controller.append_obj_log(
        logs, is_maximize=True, obj_value_is_valid=True, obj_bound_is_valid=True
    )

    assert controller.obj_value_log == [(0.0, 10.0), (1.0, 15.0)]
    assert controller.obj_bound_log == [(0.0, 20.0), (1.0, 18.0)]


def test_solve_cp_model_triggers_logging():
    controller = DummyCpController(
        instance=DummyProblem(),
        shared_param_dict={},
        cp_model_class=FakeCpModel,
        subroutine_flow=DynamicDataObject({}),
        stopping_criteria=DummyStoppingCriteria({}),
    )

    output = controller.solve_current_cp_model(
        computational_time=10.0,
        num_workers=1,
        obj_value_is_valid=True,
        obj_bound_is_valid=True,
    )

    assert output.status == "OPTIMAL"
    assert output.elapsed_time == 3.0
    assert output.objective_value == 85.0
    assert output.best_objective_bound == 50.0

    assert controller.obj_value_log[-1] == (2.0, 85.0)
    assert controller.obj_bound_log[-1] == (2.0, 70.0)



from ortools.sat.python import cp_model

from mbls import ElapsedTimer
from mbls.cpsat import SolutionProgressLogger


def test_solution_progress_logger_records_solutions():
    # Arrange
    model = cp_model.CpModel()
    x = model.NewIntVar(0, 2, "x")
    y = model.NewIntVar(0, 2, "y")
    model.Add(x != y)
    model.Maximize(x + y)

    timer = ElapsedTimer()
    logger = SolutionProgressLogger(
        elapsed_timer=timer, print_on_solution_callback=False
    )

    # Act
    solver = cp_model.CpSolver()
    solver.parameters.enumerate_all_solutions = True
    solver.solve(model, logger)
    log = logger.get_log()

    # Assert
    assert isinstance(log, list)
    assert len(log) >= 1  # At least one solution is found
    for entry in log:
        elapsed, objective, best_bound = entry
        assert isinstance(elapsed, float)
        assert isinstance(objective, float)
        assert isinstance(best_bound, float)
    # Optionally: Check the timer increments as expected
    elapsed_times = [entry[0] for entry in log]
    assert elapsed_times == sorted(elapsed_times)  # should be increasing



import warnings
from typing import Optional

from google.protobuf.internal.containers import RepeatedCompositeFieldContainer
from ortools.sat.cp_model_pb2 import ConstraintProto, CpSolverStatus
from ortools.sat.python.cp_model import CpModel, CpSolver, IntVar

from .. import ElapsedTimer
from .solution_progress_logger import SolutionProgressLogger
from .status import CpSatStatus


class CustomCpModel(CpModel):
    """A custom CpModel class that extends the OR-Tools CpModel class."""

    solver: CpSolver
    """CpSolver object for solving the model."""
    sol_prog_logger: SolutionProgressLogger
    """(Optional) Logger for solution progress."""

    num_base_constraints: int
    """Number of base constraints in the model."""

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.num_base_constraints = 0

    def solve_and_get_status(
        self, computational_time: float, num_workers: int
    ) -> tuple[CpSolverStatus, float, float, float]:
        """Solve the CP model.

        Args:
            computational_time (float): The maximum computational time in seconds.
            num_workers (int): The number of parallel workers (i.e. threads) to use during search.

        Returns:
            tuple[CpSolverStatus, float, float, float]: A tuple containing
            - the solver status,
            - elapsed time,
            - the upper bound of the objective function, and
            - the lower bound of the objective function.
        """  # noqa: E501
        self.init_solver(computational_time, num_workers)

        solver_status = self.solver.solve(self)
        elapsed_time = self.solver.wall_time

        if CpSatStatus.found_feasible_solution(solver_status):
            ub = self.solver.objective_value
            lb = self.solver.best_objective_bound
        else:
            ub, lb = CpSatStatus.get_obj_value_and_bound_for_infeasible(
                self.is_maximize()
            )

        return solver_status, elapsed_time, ub, lb

    def init_solver(
        self,
        computational_time: float,
        num_workers: int,
        random_seed: Optional[int] = None,
    ) -> None:
        """Initializes the solver.

        Args:
            computational_time (float): The maximum computational time in seconds.
            num_workers (int): The number of parallel workers (i.e. threads) to use during search.
            random_seed (Optional[int], optional): Random seed for reproducibility. Defaults to None.
        """  # noqa: E501
        self.solver = CpSolver()
        self.solver.parameters.max_time_in_seconds = computational_time
        self.solver.parameters.num_workers = num_workers
        if random_seed is not None:
            self.solver.parameters.random_seed = random_seed

    def solve_with_prog_logger(
        self,
        computational_time: float,
        num_workers: int,
        random_seed: Optional[int] = None,
        timer: Optional[ElapsedTimer] = None,
    ) -> tuple[str, float, float, float]:
        """Solve the CP model with a solution progress logger.

        Args:
            computational_time (float): The maximum computational time in seconds.
            num_workers (int): The number of parallel workers (i.e. threads) to use during search.
            random_seed (Optional[int], optional): Random seed for reproducibility. Defaults to None.
            timer (Optional[ElapsedTimer], optional): Timer to be passed to solver callback. Defaults to None.

        Returns:
            tuple[str, float, float, float]: A tuple containing
            - the solver status as a string defined in SolverStatus,
            - elapsed time in seconds,
            - the upper bound of the objective function, and
            - the lower bound of the objective function.
        """  # noqa: E501
        self.init_solver(computational_time, num_workers, random_seed)
        self.sol_prog_logger = SolutionProgressLogger(
            timer, log_info_on_solution_callback=True
        )

        solver_status = self.solver.solve(self, solution_callback=self.sol_prog_logger)
        elapsed_time = self.solver.wall_time

        if CpSatStatus.found_feasible_solution(solver_status):
            obj_value = self.solver.objective_value
            obj_bound = self.solver.best_objective_bound
        else:
            obj_value, obj_bound = CpSatStatus.get_obj_value_and_bound_for_infeasible(
                self.is_maximize()
            )

        return (
            CpSatStatus.get_status_string(solver_status),
            elapsed_time,
            obj_value,
            obj_bound,
        )

    def get_progress_log(self) -> list[tuple[float, float, float]]:
        """Returns the progress log list.

        Returns:
            list[tuple[float, float, float]]: a list of tuples
                containing (elapsed time, objective value, best bound)
        """
        return self.sol_prog_logger.get_log() if self.sol_prog_logger else []

    # variable functions

    def change_domain(self, var: IntVar, domain: list[int]) -> None:
        """Changes the domain of a variable.

        Args:
            var (IntVar)
            domain (list[int]): A list of two integers representing the new domain.
        """
        assert len(domain) == 2, (
            f"Domain must be a list of two integers; {domain} given."
        )

        var.Proto().domain[:] = domain

    # objective functions

    def is_maximize(self) -> bool:
        """
        Returns:
            bool: True if the objective is maximize, False if minimize.
        """
        proto = self.Proto()
        # If the objective is not set to maximize, it defaults to minimize
        # in OR-Tools, so we return False if maximize is not set.
        return getattr(proto.objective, "maximize", False)

    # constraint functions

    def _get_constraints(self) -> RepeatedCompositeFieldContainer[ConstraintProto]:
        proto = self.Proto()
        if not hasattr(proto, "constraints") or proto.constraints is None:
            raise RuntimeError("No constraints defined in the model.")
        return proto.constraints

    def get_next_constr_idx(self) -> int:
        """Returns the index of the next constraint.

        Returns:
            int: The index of the next constraint.
        """
        return len(self._get_constraints())

    def set_num_base_constraints(self) -> None:
        """Sets the base number of constraints to the current number of constraints."""
        self.num_base_constraints = self.get_next_constr_idx()

    def freeze_base_constraints(self) -> None:
        warnings.warn(
            "freeze_base_constraints() is deprecated and will be removed in a future version. "
            "Use set_num_base_constraints() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.set_num_base_constraints()

    # methods to delete constraints

    def delete_constraints(self, idx_start: int, idx_end: int) -> None:
        del self._get_constraints()[idx_start:idx_end]

    def delete_added_constraints(self):
        """Deletes all constraints added after base model was built.

        Raises:
            ValueError: If no constraints were added after the base model was built.
        """

        if self.num_base_constraints == 0:
            raise ValueError("No base model constraints defined.")
        current_num_constraints = self.get_next_constr_idx()
        if current_num_constraints > self.num_base_constraints:
            self.delete_constraints(self.num_base_constraints, current_num_constraints)



import logging
from typing import Optional

from ortools.sat.python import cp_model

from .. import ElapsedTimer


class SolutionProgressLogger(cp_model.CpSolverSolutionCallback):
    """
    A lightweight logger for tracking the progress of solution discovery
    during CP-SAT solving.

    This class is designed to be used as a callback with the OR-Tools CpSolver.
    It logs a timestamped record of each feasible solution found, capturing:

    - Elapsed time since solving began
    - Objective value of the current solution
    - Best known lower (upper) bound found when min(max)imizing at that moment

    The logger does **not** analyze or interpret results.
    It only accumulates raw progress data, which can be retrieved later
    (e.g., for plotting, reporting, or summary generation).

    Example usage:
        >>> logger = SolutionProgressLogger(timer)
        >>> solver.SolveWithSolutionCallback(model, logger)
        >>> progress = logger.get_log()
    """

    _e_timer: ElapsedTimer
    _log: list[tuple[float, float, float]]
    """A list of tuples containing (elapsed time, objective value, best bound)."""

    def __init__(
        self,
        elapsed_timer: Optional[ElapsedTimer] = None,
        print_on_solution_callback: bool = False,
        log_info_on_solution_callback: bool = False,
    ) -> None:
        super().__init__()
        self._log = []
        if elapsed_timer is None:
            self._e_timer = ElapsedTimer()
            self._e_timer.set_start_time_as_now()
        else:
            self._e_timer = elapsed_timer
        self._print_on_solution_callback = print_on_solution_callback
        """If True, prints progress on each solution callback."""
        self._log_info_on_solution_callback = log_info_on_solution_callback
        """If True, logs additional information on each solution callback."""

    def on_solution_callback(self) -> None:
        """
        Callback method called on each solution found.

        - The method must be implemented in subclass of cp_model.CpSolverSolutionCallback.
        - This method logs the elapsed time, objective value, and best bound
        at the time of the solution discovery.
        - It also prints the progress if `self._print_on_solution_callback` is True.
        - It logs additional information if `self._log_info_on_solution_callback` is True.
        """
        elapsed = self._e_timer.get_elapsed_sec()
        objective = self.objective_value
        best_bound = self.best_objective_bound
        self._log.append((elapsed, objective, best_bound))
        info_str = (
            f"Elapsed: {elapsed:.2f} sec, "
            f"Objective: {objective}, "
            f"Best Bound: {best_bound}"
        )
        if self._print_on_solution_callback:
            print(info_str)
        if self._log_info_on_solution_callback:
            logging.info(info_str)

    @property
    def log(self) -> list[tuple[float, float, float]]:
        """Property to access the log list.

        Returns:
            list[tuple[float, float, float]]: a list of tuples
                containing (elapsed time, objective value, best bound)
        """
        return self._log.copy()

    def get_log(self) -> list[tuple[float, float, float]]:
        """Returns the log list.

        Returns:
            list[tuple[float, float, float]]: a list of tuples
                containing (elapsed time, objective value, best bound)
        """
        return self.log



# routix classes
from routix import (
    DynamicDataObject,
    ElapsedTimer,
    ExperimentSummary,
    SolverOutputSummary,
    SolverStatus,
    SubroutineController,
    SubroutineFlowValidator,
    utils,
)

__all__ = [
    "DynamicDataObject",
    "ElapsedTimer",
    "ExperimentSummary",
    "SolverOutputSummary",
    "SolverStatus",
    "SubroutineController",
    "SubroutineFlowValidator",
    "utils",
]



from __future__ import annotations

import warnings
from pathlib import Path
from typing import Any

import yaml
from routix.metric_time_series import MetricTimeSeries, NamedTimeSeriesStore
from routix.type_hints import Numeric


class ObjValueBoundStore(NamedTimeSeriesStore[Numeric]):
    """A store for objective value and bound time series data.

    This class extends NamedTimeSeriesStore to specifically handle
    objective values and bounds in a time series format.
    """

    def __init__(self) -> None:
        super().__init__()
        self.obj_value_series: MetricTimeSeries[Numeric] = self.get_or_create(
            "obj_value"
        )
        """MetricTimeSeries for objective values."""
        self.obj_bound_series: MetricTimeSeries[Numeric] = self.get_or_create(
            "obj_bound"
        )
        """MetricTimeSeries for objective bounds."""

    def add_obj_value(
        self, timestamp: float, value: Numeric, is_maximize: bool = False
    ) -> None:
        if is_maximize:
            self.obj_value_series.add_if_value_stg_last(timestamp, value)
        else:
            self.obj_value_series.add_if_value_stl_last(timestamp, value)

    def add_obj_bound(
        self, timestamp: float, value: Numeric, is_maximize: bool = False
    ) -> None:
        if is_maximize:
            self.obj_bound_series.add_if_value_stl_last(timestamp, value)
        else:
            self.obj_bound_series.add_if_value_stg_last(timestamp, value)

    def add_last_timestamp_note(
        self,
        note: Any,
        do_overwrite: bool = False,
        obj_value_is_valid: bool = False,
        obj_bound_is_valid: bool = False,
    ) -> None:
        """
        Add a note to the last timestamp of the objective value and bound series.
        This method checks the last timestamps of both the objective value and bound
        series to determine where to add the note. If both series have valid last
        timestamps, the note is added to both. If only one series has a valid last
        timestamp, the note is added to that series.

        Args:
            note (Any): Note to be added to the last timestamp.
            overwrite (bool, optional): If True, overwrite the last value with the note.
                Defaults to False.
            obj_value_is_valid (bool, optional): True if the last objective value is valid.
                Defaults to False.
            obj_bound_is_valid (bool, optional): True if the last objective bound is valid.
                Defaults to False.
        """
        obj_value_last_timestamp: float | None = self.obj_value_series.last_timestamp
        obj_bound_last_timestamp: float | None = self.obj_bound_series.last_timestamp

        obj_value_note_cond = (
            obj_value_is_valid and obj_value_last_timestamp is not None
        )
        obj_bound_note_cond = (
            obj_bound_is_valid and obj_bound_last_timestamp is not None
        )
        if not (obj_value_note_cond or obj_bound_note_cond):
            warnings.warn("No valid timestamp found for appending note.")
            return
        # Determine the last timestamp to use for adding the note
        # If both are None, we have no data to add a note to
        if obj_value_note_cond and obj_bound_note_cond:
            # If both are valid, we use the maximum of the two
            assert obj_value_last_timestamp is not None
            assert obj_bound_last_timestamp is not None
            last_timestamp = max(obj_value_last_timestamp, obj_bound_last_timestamp)
            self.obj_value_series.repeat_last_value(
                last_timestamp, note=note, overwrite_note=do_overwrite
            )
            self.obj_bound_series.repeat_last_value(
                last_timestamp, note=note, overwrite_note=do_overwrite
            )
        # If one is None, we use the other
        elif obj_value_note_cond:
            assert obj_value_last_timestamp is not None
            self.obj_value_series.repeat_last_value(
                obj_value_last_timestamp, note=note, overwrite_note=do_overwrite
            )
        elif obj_bound_note_cond:
            assert obj_bound_last_timestamp is not None
            self.obj_bound_series.repeat_last_value(
                obj_bound_last_timestamp, note=note, overwrite_note=do_overwrite
            )

    # I/O from/to dict

    def to_dict(self) -> dict[str, dict[str, Any]]:
        return super().to_dict()

    @classmethod
    def from_dict(cls, data: dict[str, dict[str, Any]]) -> ObjValueBoundStore:
        """Create an ObjValueBoundStore from a dictionary."""
        # obj_value_series and obj_bound_series will be initialized in __init__
        store = cls()
        store.clear()  # Clear existing data before loading new data

        for name, ts_data in data.items():
            store._store[name] = MetricTimeSeries.from_dict(ts_data)

        # Ensure obj_value_series and obj_bound_series are initialized or binded
        store.obj_value_series = store.get_or_create("obj_value")
        store.obj_bound_series = store.get_or_create("obj_bound")
        return store

    # I/O from/to YAML

    def save_yaml(self, file_path: Path | str, encoding: str = "utf-8") -> None:
        """Save to a YAML file.

        Args:
            file_path (str): Path to the YAML file where the store will be saved.
            encoding (str, optional): Encoding to use when writing the file. Defaults to "utf-8".
        """
        super().save_yaml(file_path, encoding=encoding)

    @classmethod
    def load_yaml(
        cls, file_path: Path | str, encoding: str = "utf-8"
    ) -> ObjValueBoundStore:
        """Load an ObjValueBoundStore from a YAML file.

        Args:
            file_path (str): Path to the YAML file from which the store will be loaded.
            encoding (str, optional): Encoding to use when writing the file. Defaults to "utf-8".

        Returns:
            ObjValueBoundStore[Numeric]: An instance of ObjValueBoundStore populated with the data.
        """
        with open(file_path, "r") as file:
            data = yaml.safe_load(file)
        return cls.from_dict(data)



from ortools.sat.cp_model_pb2 import CpSolverStatus
from ortools.sat.python.cp_model import (
    FEASIBLE,
    INFEASIBLE,
    INT_MAX,
    INT_MIN,
    MODEL_INVALID,
    OPTIMAL,
    UNKNOWN,
)

from .. import SolverStatus


class CpSatStatus:
    cpsat_status_dict = {
        UNKNOWN: SolverStatus.UNKNOWN,
        MODEL_INVALID: SolverStatus.MODEL_INVALID,
        FEASIBLE: SolverStatus.FEASIBLE,
        INFEASIBLE: SolverStatus.INFEASIBLE,
        OPTIMAL: SolverStatus.OPTIMAL,
    }
    """Map: ortools.sat.python.cp_model status codes -> string"""

    feasible_cpsat_status_set = frozenset({FEASIBLE, OPTIMAL})
    """Set of status codes indicating a feasible solution was found."""

    @staticmethod
    def get_status_string(cpsat_status: CpSolverStatus) -> str:
        """Returns the status string corresponding to the given status code."""
        for key, value in CpSatStatus.cpsat_status_dict.items():
            if key == cpsat_status:
                return value
        raise ValueError(f"Invalid status code: {cpsat_status}")

    @staticmethod
    def found_feasible_solution(cpsat_status: CpSolverStatus) -> bool:
        """Checks if a feasible solution was found based on the status code."""
        return cpsat_status in CpSatStatus.feasible_cpsat_status_set

    @staticmethod
    def get_obj_value_and_bound_for_infeasible_maximize() -> tuple[int, int]:
        """Returns the upper and lower bounds for maximization problems."""
        return INT_MIN, INT_MIN

    @staticmethod
    def get_obj_value_and_bound_for_infeasible_minimize() -> tuple[int, int]:
        """Returns the upper and lower bounds for minimization problems."""
        return INT_MAX, INT_MAX

    @staticmethod
    def get_obj_value_and_bound_for_infeasible(is_maximize: bool) -> tuple[int, int]:
        """Returns the upper and lower bounds for infeasible problems."""
        if is_maximize:
            return CpSatStatus.get_obj_value_and_bound_for_infeasible_maximize()
        else:
            return CpSatStatus.get_obj_value_and_bound_for_infeasible_minimize()



from typing import Hashable

from ortools.sat.python.cp_model import IntervalVar, IntVar

from .custom_cp_model import CustomCpModel


class CpModelWithFixedInterval(CustomCpModel):
    # Parameter

    horizon: int
    """
    The horizon for the scheduling problem, which is the maximum time
    that any operation can start or end.
    This is used to define the domain of the start and end time variables.
    """

    # Variables

    var_op_start: dict[Hashable, IntVar]
    """
    Dictionary to store start time variables for each operation in a job.
    """
    var_op_end: dict[Hashable, IntVar]
    """
    Dictionary to store end time variables for each operation in a job.
    """
    var_op_intvl: dict[Hashable, IntervalVar]
    """
    Dictionary to store interval variables for each operation in a job.
    """

    # Interval variable name set
    var_op_intvl_fixed_name_set: set[str]

    def __init__(self, horizon: int):
        """Initialize the CpModelWithFixedInterval class.

        Args:
            horizon (int): The horizon for the scheduling problem,
                           which is the maximum time that any operation can end.
        """
        super().__init__()

        # Parameter
        self.horizon = horizon

        # Initialize dictionaries to store variables
        self.var_op_start = {}
        self.var_op_end = {}
        self.var_op_intvl = {}

        # Initialize the set to store interval variable names
        self.var_op_intvl_fixed_name_set = set()

    def define_fixed_interval_var(
        self, job_idx: str, stage_idx: str, mc_idx: str, processing_time: int
    ):
        """Define a fixed interval variable for an operation.

        Args:
            job_idx (str): The index of the job.
            stage_idx (str): The index of the stage.
            mc_idx (str): The index of the machine.
            processing_time (int): The processing time of the operation.
        """
        suffix = f"{job_idx}_{stage_idx}_{mc_idx}"
        var_op_intvl_fixed_name = f"intvl_fixed_{suffix}"
        # Check if the interval variable name already exists
        if var_op_intvl_fixed_name in self.var_op_intvl_fixed_name_set:
            raise ValueError(
                f"Interval variable name '{var_op_intvl_fixed_name}' already exists."
            )

        # Create the fixed interval variable with fixed processing time
        start_var = self.NewIntVar(0, self.horizon, f"start_{suffix}")
        end_var = self.NewIntVar(0, self.horizon, f"end_{suffix}")
        intvl_var = self.new_interval_var(
            start_var, processing_time, end_var, var_op_intvl_fixed_name
        )

        # Add each variable to the corresponding dictionaries
        self.var_op_start[job_idx, stage_idx, mc_idx] = start_var
        self.var_op_end[job_idx, stage_idx, mc_idx] = end_var
        self.var_op_intvl[job_idx, stage_idx, mc_idx] = intvl_var

        # Add name of the new variable to the set
        self.var_op_intvl_fixed_name_set.add(var_op_intvl_fixed_name)



from .cp_model_with_fixed_interval import CpModelWithFixedInterval
from .cp_model_with_optional_fixed_interval import CpModelWithOptionalFixedInterval
from .cp_subroutine_controller import CpSubroutineController
from .custom_cp_model import CustomCpModel
from .obj_value_bound_store import ObjValueBoundStore
from .solution_progress_logger import SolutionProgressLogger
from .status import CpSatStatus

__all__ = [
    "CpModelWithFixedInterval",
    "CpModelWithOptionalFixedInterval",
    "CpSubroutineController",
    "CustomCpModel",
    "ObjValueBoundStore",
    "SolutionProgressLogger",
    "CpSatStatus",
]
